# -*- coding: utf-8 -*-

import os
import sys
import webbrowser
import datetime
import pdb

from bingads.service_client import ServiceClient
from bingads.authorization import AuthorizationData, OAuthDesktopMobileAuthCodeGrant
from bingads.exceptions import OAuthTokenRequestException

from bing_content_api.api import BingContentApi

from shopping_utils import logger


class BingShopping(object):
    def __init__(self, merchant_id, client_id, developer_token, refresh_token, environment='production', catalog_id=None):
        self.merchant_id = merchant_id
        self.client_id = client_id
        self.developer_token = developer_token
        self.refresh_token = refresh_token
        self.environment = environment
        self.catalog_id = catalog_id

        self.max_retries = 9

        self.api = BingContentApi(
            client_id, developer_token, merchant_id, refresh_token, environment,
            catalog_id=catalog_id)

    def download_inventory(self, output_path=None, **kwargs):
        inv = dict()

        for product in self.api.Product.iterall(**kwargs):
            try:
                offer_id = product.id.split(':').pop()
            except Exception as e:
                logger.exception(e)
                logger.warning('[InvalidBingMerchantRecord] %s', repr(product))
                continue

            k = offer_id
            price = product.price
            try:
                inv[k] = {
                    'product_id': product.id,
                    'availability': product.availability,
                    'price': round(float(price['value']), 2),
                    'currency': price['currency']
                }
            except Exception as e:
                logger.warning('[InvalidBingMerchantRecord] %s', repr(product))

        if output_path:
            output_dir = os.path.dirname(os.path.abspath(os.path.expanduser(output_path)))
            if not os.path.isdir(output_dir):
                os.makedirs(output_dir)
            with open(output_path, 'w') as fh:
                headers = ['merchant_id', 'product_id', 'price', 'currency', 'availability']
                fh.write('\t'.join(headers))
                fh.write('\n')

                merchant_id = str(self.merchant_id)
                for v in inv.values():
                    record = [merchant_id]
                    for header in headers:
                        if header in v:
                            record.append(str(v[header]))
                    fh.write('\t'.join(record))
                    fh.write('\n')

        return inv

    def upload_products(self, prod_buf):
        result = {
            'succeed_cnt': 0,
            'failed_cnt': 0
        }

        if not prod_buf:
            return result

        batch = {'entries': []}
        for idx, item in enumerate(prod_buf):
            entry = {
                'batchId': idx,
                'merchantId': self.merchant_id,
                'method': 'insert',
                'product': item
            }
            batch['entries'].append(entry)

        resp = self.api.Product.batch(batch)
        for entry in resp.get('entries', []):
            if 'errors' in entry and entry['errors']:
                result['failed_cnt'] += 1
                if 'product' in entry and entry['product']:
                    logger.error(
                        '[CreateProductFailed] OfferId: %s, errors: %s',
                        entry['product']['offerId'], repr(entry['errors']))
                else:
                    logger.warning(
                        '[CreateProductFailed] Entry: %s', repr(entry))
            else:
                result['succeed_cnt'] += 1
                if 'product' in entry and entry['product']:
                    logger.info('[ProductCreated] offerId %s!', entry['product']['offerId'])
                else:
                    logger.info(
                        '[ProductCreated] %s, but no product information return!', repr(entry))

        return result

    def update_inventory(self, inv_buf):
        result = {'succeed_cnt': 0, 'failed_cnt': 0}
        batch = {'entries': []}
        for idx, item in enumerate(inv_buf):
            productId = item.pop('productId')
            entry = {
                'batchId': idx,
                'merchantId': self.merchant_id,
                'productId': productId,
                'inventory': item
            }
            batch['entries'].append(entry)

        resp = self.api.Inventory.batch(batch)
        for entry in resp['entries']:
            errors = entry.get('errors')
            if errors:
                result['failed_cnt'] += 1
                logger.error(repr(errors))
            else:
                result['succeed_cnt'] += 1

                try:
                    prod = batch['entries'][int(entry['batchId'])]
                    logger.info(
                        "[MerchantId: %s][ProductId: %s] availability - %s, price - %s",
                        prod['merchantId'], prod['productId'],
                        prod['inventory']['availability'],
                        str(prod['inventory'].get('price', '-')))
                except Exception as e:
                    logger.exception(e)

        return result

    def update_expiration(self, days_before_expire=14):
        days = datetime.timedelta(days=30)
        now = datetime.datetime.now()
        expiration_date = now + days
        need_update_date = now + datetime.timedelta(days=days_before_expire)
        expire_str = expiration_date.strftime("%Y-%m-%dT%H:%M:%SZ")
        default_date_str = now.strftime("%Y-%m-%dT%H:%M:%SZ")

        result = {
            'processed_cnt': 0,
            'active_cnt': 0,
            'expiring_cnt': 0,
            'expired_cnt': 0,
            'succeed_cnt': 0,
            'failed_cnt': 0,
        }
        batch = {'entries': []}
        idx = 0
        batch_size = 250
        product_ids = []
        for productstatus in self.list_product_statuses():
            cur_expiration_date = None
            cur_expiration_date_s = productstatus['expirationDate'].split('.').pop(0)

            result['processed_cnt'] += 1

            expiration_date_invalid = False
            try:
                cur_expiration_date = datetime.datetime.strptime(
                    cur_expiration_date_s, "%Y-%m-%dT%H:%M:%S")
                if cur_expiration_date <= now:
                    result['expired_cnt'] += 1
            except ValueError as e:
                logger.exception(e)
                expiration_date_invalid = True

            if (cur_expiration_date and cur_expiration_date <= need_update_date) or expiration_date_invalid:
                if not expiration_date_invalid:
                    result['expiring_cnt'] += 1

                product_ids.append(productstatus['productId'])
            else:
                result['active_cnt'] += 1

            if len(product_ids) < batch_size:
                continue

            res = self.update_products_expiration(product_ids, expire_str)
            result['succeed_cnt'] += res['succeed_cnt']
            result['failed_cnt'] += res['failed_cnt']
            product_ids = []

        if product_ids:
            res = self.update_products_expiration(product_ids, expire_str)
            result['succeed_cnt'] += res['succeed_cnt']
            result['failed_cnt'] += res['failed_cnt']

        return result

    def cleanup(self):
        result = {
            'succeed_cnt': 0,
            'failed_cnt': 0
        }

        product_ids = []
        buf_size = 250
        for productstatus in self.list_product_statuses():
            if productstatus['status'] != 'Disapproved':
                continue

            product_ids.append(productstatus['productId'])
            if len(product_ids) < buf_size:
                continue

            res = self.delete_products(product_ids)
            result['succeed_cnt'] += res['succeed_cnt']
            result['failed_cnt'] += res['failed_cnt']

            product_ids = []

        if product_ids:
            res = self.delete_products(product_ids)
            result['succeed_cnt'] += res['succeed_cnt']
            result['failed_cnt'] += res['failed_cnt']

            product_ids = []

        return result

    def delete_products(self, productids):
        result = {
            'succeed_cnt': 0,
            'failed_cnt': 0
        }

        batch = {'entries': []}
        for idx, productid in enumerate(productids):
            entry = {
                'batchId': idx,
                'merchantId': self.merchant_id,
                'method': 'delete',
                'productId': productid
            }
            batch['entries'].append(entry)

        resp = self.api.Product.batch(batch)

        for entry in resp['entries']:
            if 'errors' in entry and entry['errors']:
                result['failed_cnt'] += 1
                logger.warning(
                    '[ProductDeleteFailed] ProductID: %s, Errors: %s',
                    batch['entries'][int(entry['batchId'])]['productId'], entry['errors'])
            else:
                result['succeed_cnt'] += 1
                logger.info(
                    '[ProductDeleted] ProductID: %s',
                    batch['entries'][int(entry['batchId'])]['productId'])

        return result

    def update_products_expiration(self, product_ids, expire_str):
        result = {
            'succeed_cnt': 0,
            'failed_cnt': 0
        }
        expiration_update_batch = {'entries': []}
        expiration_update_entry_idx = 0

        batch = {'entries': []}
        for idx, product_id in enumerate(product_ids):
            batch['entries'].append({
                'batchId': idx,
                'merchantId': self.merchant_id,
                'method': 'get',
                'productId': product_id,
            })
        resp = self.api.Product.batch(batch)
        for entry in resp.get('entries', []):
            if 'errors' in entry and entry['errors']:
                result['failed_cnt'] += 1
                logger.warning(
                    'Could not get product information! Entry: %s', repr(entry))
                continue

            product = entry['product']
            if 'source' in product:
                product.pop('source')

            product['expirationDate'] = expire_str
            expiration_update_batch['entries'].append({
                'batchId': expiration_update_entry_idx,
                'merchantId': self.merchant_id,
                'method': 'insert',
                'product': product
            })
            expiration_update_entry_idx += 1

        if not expiration_update_batch['entries']:
            return result

        resp = self.api.Product.batch(expiration_update_batch)
        for entry in resp.get('entries', []):
            if 'errors' in entry and entry['errors']:
                result['failed_cnt'] += 1
                if 'product' in entry and entry['product']:
                    logger.error(
                        'Expiration date update failed! OfferId: %s, errors: %s',
                        entry['product']['offerId'], repr(entry['errors']))
                else:
                    logger.warning(
                        'Expiration date update failed! Entry: %s', repr(entry))
            else:
                result['succeed_cnt'] += 1
                if 'product' in entry and entry['product']:
                    logger.info(
                        'Expiration date of product with offerId %s is updated to %s!',
                        entry['product']['offerId'], expire_str)
                else:
                    logger.info(
                        'Expiration date updated to %s,'
                        ' but no product information return!'
                        'Entry: %s',
                        expire_str, repr(entry))

        return result

    def list_product_statuses(self, **kwargs):
        for productstatus in self.api.ProductStatus.iterall(**kwargs):
            yield productstatus

    def authenticate(self):
        self.api.connection.authenticate()

    def get_refresh_token(self, callback=None):
        auth_data = AuthorizationData(
            account_id=None,
            customer_id=None,
            developer_token=self.developer_token,
            authentication=None,
        )

        auth = OAuthDesktopMobileAuthCodeGrant(
            client_id=self.client_id,
            env=self.environment
        )

        auth.state = self.api.connection.state

        # Assign this authentication instance to the authorization_data. 
        auth_data.authentication = auth

        # Register the callback function to automatically save the refresh token anytime it is refreshed.
        if callback:
            auth_data.authentication.token_refreshed_callback = callback

        token = None
        try:
            # If we have a refresh token let's refresh it
            if self.refresh_token is not None:
                token = auth_data.authentication.request_oauth_tokens_by_refresh_token(self.refresh_token)
            else:
                token = self.request_user_consent(auth_data)
        except OAuthTokenRequestException:
            # The user could not be authenticated or the grant is expired. 
            # The user must first sign in and if needed grant the client application access to the requested scope.
            toekn = self.request_user_consent(auth_data)

        return token

    def request_user_consent(self, auth_data):
        webbrowser.open(auth_data.authentication.get_authorization_endpoint(), new=1)
        # For Python 3.x use 'input' instead of 'raw_input'
        if(sys.version_info.major >= 3):
            response_uri = input(
                "You need to provide consent for the application to access your Microsoft Advertising accounts. " \
                "After you have granted consent in the web browser for the application to access your Microsoft Advertising accounts, " \
                "please enter the response URI that includes the authorization 'code' parameter: \n"
            )
        else:
            response_uri = raw_input(
                "You need to provide consent for the application to access your Microsoft Advertising accounts. " \
                "After you have granted consent in the web browser for the application to access your Microsoft Advertising accounts, " \
                "please enter the response URI that includes the authorization 'code' parameter: \n"
            )

        if auth_data.authentication.state != self.api.connection.state:
            raise Exception("The OAuth response state does not match the client request state.")

        # Request access and refresh tokens using the URI that you provided manually during program execution.
        return auth_data.authentication.request_oauth_tokens_by_response_uri(response_uri=response_uri) 
