# -*- coding: utf-8 -*-

import os
import time
import datetime
from socket import error as SocketError
import errno

from google.oauth2 import service_account
import google_auth_httplib2
from googleapiclient import discovery, http
from googleapiclient.errors import HttpError

from shopping_utils import logger, APPLICATION_NAME


SERVICE_NAME = 'content'
SERVICE_VERSION = 'v2.1'
SANDBOX_SERVICE_VERSION = 'v2sandbox'
CONTENT_API_SCOPE = 'https://www.googleapis.com/auth/' + SERVICE_NAME


class GoogleShopping(object):
    def __init__(self, merchant_id, service_account_info, service_version=SERVICE_VERSION):
        self.service_version = service_version
        self.merchant_id = merchant_id
        self.service_account = service_account_info
        try:
            self.service = self.authenticate(self.service_account, self.service_version)
        except HttpError as e:
            if e.resp.status == 401:
                logger.error('[GoogleMerchantUnauthorized] %s', repr(e))
                msg = 'Could not authorize or authenticate google merchant '
                msg += 'with given service account - '
                msg += service_account_info
                raise ValueError(msg)

        self.max_retries = 9

    def download_inventory(self, output_path=None, callbacks=None):
        inv = dict()

        page_size = 250
        request = self.service.products().list(
            merchantId=self.merchant_id, maxResults=page_size)

        i = 0
        while request is not None:
            result = request.execute(num_retries=self.max_retries)
            if 'resources' not in result or not result['resources']:
                logger.debug("Google merchant inventory downloaded!")
                break

            i += 1
            logger.debug("Downloaded google merchant inventory page %d", i)

            for prod in result['resources']:
                product_id = prod['id']
                k = product_id.split(':')[-1]

                custom_attributes = dict()
                for attr in prod.get('customAttributes', []):
                    attr_name = attr['name']
                    if attr_name == 'price':
                        custom_attributes[attr_name] = round(float(attr['value']), 2)
                        custom_attributes['currency'] = attr.get('unit', 'USD')
                    else:
                        custom_attributes[attr_name] = attr['value']
                try:
                    availability = prod.get('availability', custom_attributes.get('availability'))
                    if 'price' in prod:
                        price = round(float(prod['price']['value']), 2)
                        currency = prod['price']['currency']
                    else:
                        price = custom_attributes['price']
                        currency = custom_attributes['currency']
                    inv[k] = {
                        'product_id': product_id,
                        'availability': availability,
                        'price': price,
                        'currency': currency
                    }
                except Exception as e:
                    logger.warning('[InvalidGoogleMerchantRecord] %s', repr(prod))

            if callbacks:
                for callback in callbacks:
                    if not callable(callback):
                        continue

                    try:
                        callback(result['resources'])
                    except Exception as e:
                        logger.exception(e)

            request = self.service.products().list_next(request, result)

        if output_path:
            output_dir = os.path.dirname(output_path)
            if not os.path.isdir(output_dir):
                os.makedirs(output_dir)
            with open(output_path, 'w') as fh:
                headers = ['merchant_id', 'product_id', 'price', 'currency', 'availability']
                fh.write('\t'.join(headers))
                fh.write('\n')

                merchant_id = str(self.merchant_id)
                for v in inv.values():
                    record = [merchant_id]
                    for header in headers:
                        if header in v:
                            record.append(str(v[header]))
                    fh.write('\t'.join(record))
                    fh.write('\n')

        return inv

    def download_statuses(self):
        page_size = 250
        request = self.service.productstatuses().list(
            merchantId=self.merchant_id, maxResults=page_size)

        i = 0
        while request is not None:
            resp = request.execute(num_retries=self.max_retries)
            request = self.service.productstatuses().list_next(request, resp)

            i += 1
            logger.debug("Downloaded merchant product status page %d" % i)

            prodstatus_buf = []
            if 'resources' in resp and resp['resources']:
                for prodstatus in resp['resources']:
                    if 'destinationStatuses' not in prodstatus:
                        logger.debug("No destinationStatuses found: %s", repr(prodstatus))
                        continue

                    shopping_stat = next(
                        (dest_stat for dest_stat in prodstatus['destinationStatuses'] \
                            if dest_stat['destination'] == 'Shopping'), None)
                    status = shopping_stat.get('approvalStatus', shopping_stat.get('status', None)) \
                        if shopping_stat else None
                    if status is None:
                        logger.warning('[MissingStatus] %s', prodstatus)
                        continue

                    # if 'dataQualityIssues' not in prodstatus and \
                    #     'itemLevelIssues' not in prodstatus:
                    #     logger.debug(
                    #         'Neither dataQualityIssues nor itemLevelIssues find in product,'
                    #         ' skip: %s',
                    #         repr(prodstatus))
                    #     continue

                    yield prodstatus 

    def get_product_statuses(self, product_ids):
        product_statuses = dict()

        batch = {'entries': []}
        for idx, product_id in enumerate(product_ids):
            batch['entries'].append({
                'batchId': idx,
                'merchantId': self.merchant_id,
                'method': 'get',
                'productId': product_id
            })

        req = self.service.productstatuses().custombatch(body=batch)
        resp = req.execute(num_retries=self.max_retries)
        for entry in resp.get('entries', []):
            if 'errors' in entry and entry['errors']:
                logger.warning(
                    'Could not get product status information! Entry: %s', repr(entry))
                continue

            product_status = entry['productStatus']

            product_statuses[product_status['productId']] = product_status

        return product_statuses

    def list_products(self):
        page_size = 250
        request = self.service.products().list(
            merchantId=self.merchant_id, maxResults=page_size)

        i = 0
        while request is not None:
            result = request.execute(num_retries=self.max_retries)
            if 'resources' not in result or not result['resources']:
                logger.debug("Google merchant list products finished!")
                break

            i += 1
            logger.debug("Listed google merchant products page %d", i)

            for prod in result['resources']:
                yield prod

            request = self.service.products().list_next(request, result)

    def update_expiration(self, days_before_expire=14):
        days = datetime.timedelta(days=30)
        now = datetime.datetime.now()
        expiration_date = now + days
        need_update_date = now + datetime.timedelta(days=days_before_expire)
        expire_str = expiration_date.strftime("%Y-%m-%dT%H:%M:%SZ")
        default_date_str = now.strftime("%Y-%m-%dT%H:%M:%SZ")

        page_size = 250
        request = self.service.productstatuses().list(
            merchantId=self.merchant_id, maxResults=page_size)

        result = {
            'processed_cnt': 0,
            'active_cnt': 0,
            'expiring_cnt': 0,
            'expired_cnt': 0,
            'succeed_cnt': 0,
            'failed_cnt': 0,
        }
        i = 0
        while request is not None:
            resp = request.execute(num_retries=self.max_retries)
            request = self.service.productstatuses().list_next(request, resp)

            if 'resources' not in resp or not resp['resources']:
                break

            i += 1
            logger.debug("Downloaded merchant product status page %d" % i)

            statuses = {
                'active': [],
                'expiring': [],
                'expired': [],
            }
            batch = {'entries': []}
            for idx, prodstatus in enumerate(resp['resources']):
                logger.debug(
                    'ProductId: %s, ExpirationDate: %s',
                    prodstatus['productId'], prodstatus['googleExpirationDate'])

                result['processed_cnt'] += 1

                item = {'productId': prodstatus['productId']}
                expiration_date_invalid = False
                try:
                    cur_expiration_date = datetime.datetime.strptime(
                        prodstatus.get('googleExpirationDate', default_date_str),
                        "%Y-%m-%dT%H:%M:%SZ")
                    item['expirationDate'] = cur_expiration_date

                    if cur_expiration_date <= now:
                        result['expired_cnt'] += 1
                        statuses['expired'].append(item)
                except ValueError as e:
                    logger.exception(e)
                    expiration_date_invalid = True
                    item['expirationDate'] = None

                if cur_expiration_date <= need_update_date or expiration_date_invalid:
                    if cur_expiration_date <= need_update_date:
                        result['expiring_cnt'] += 1
                        statuses['expiring'].append(item)

                    batch['entries'].append({
                        'batchId': idx,
                        'merchantId': self.merchant_id,
                        'method': 'get',
                        'productId': prodstatus['productId']
                    })
                else:
                    result['active_cnt'] += 1
                    statuses['active'].append(item)

            # yield statuses

            if not batch['entries']:
                continue

            expiration_update_batch = {'entries': []}
            expiration_update_entry_idx = 0

            req = self.service.products().custombatch(body=batch)
            resp = req.execute(num_retries=self.max_retries)
            for entry in resp.get('entries', []):
                if 'errors' in entry and entry['errors']:
                    result['failed_cnt'] += 1
                    logger.warning(
                        'Could not get product information! Entry: %s', repr(entry))
                    continue

                product = entry['product']
                if 'source' in product:
                    product.pop('source')
                product['expirationDate'] = expire_str
                expiration_update_batch['entries'].append({
                    'batchId': expiration_update_entry_idx,
                    'merchantId': self.merchant_id,
                    'method': 'insert',
                    'product': product
                })
                expiration_update_entry_idx += 1

            if not expiration_update_batch['entries']:
                continue

            req = self.service.products().custombatch(body=expiration_update_batch)
            resp = req.execute(num_retries=self.max_retries)
            if resp['kind'] != 'content#productsCustomBatchResponse':
                result['failed_cnt'] += len(resp['entries'])
                logger.warning("Unknown kind - %s", resp['kind'])
                continue

            for entry in resp['entries']:
                if 'errors' in entry and entry['errors']:
                    result['failed_cnt'] += 1

                    if 'product' in entry and entry['product']:
                        logger.error(
                            'Expiration date update failed! OfferId: %s, errors: %s',
                            entry['product']['offerId'], repr(entry['errors']))
                    else:
                        logger.warning(
                            'Expiration date update failed! Entry: %s', repr(entry))
                else:
                    result['succeed_cnt'] += 1

                    if 'product' in entry and entry['product']:
                        logger.info(
                            'Expiration date of product with offerId %s is updated to %s!',
                            entry['product']['offerId'], expire_str)
                    else:
                        logger.info(
                            'Expiration date updated to %s,'
                            ' but no product information return!'
                            'Entry: %s',
                            expire_str, repr(entry))

        return result

    def get_products(self, product_ids):
        products = dict()

        batch = {'entries': []}
        for idx, product_id in enumerate(product_ids):
            batch['entries'].append({
                'batchId': idx,
                'merchantId': self.merchant_id,
                'method': 'get',
                'productId': product_id
            })

        req = self.service.products().custombatch(body=batch)
        resp = req.execute(num_retries=self.max_retries)
        for entry in resp.get('entries', []):
            if 'errors' in entry and entry['errors']:
                logger.warning(
                    'Could not get product information! Entry: %s', repr(entry))
                continue

            product = entry['product']
            if 'source' in product:
                product.pop('source')

            products[product['id']] = product

        return products

    def fetch_and_update_products(self, data):
        result = {
            'succeed_cnt': 0,
            'failed_cnt': 0,
        }
        batch = {'entries': []}
        for idx, product_id in enumerate(data.keys()):
            batch['entries'].append({
                'batchId': idx,
                'merchantId': self.merchant_id,
                'method': 'get',
                'productId': product_id
            })

        req = self.service.products().custombatch(body=batch)
        resp = req.execute(num_retries=self.max_retries)
        product_update_entry_idx = 0
        product_update_batch = {'entries': []}
        for entry in resp.get('entries', []):
            if 'errors' in entry and entry['errors']:
                result['failed_cnt'] += 1
                logger.warning(
                    'Could not get product information! Entry: %s', repr(entry))
                continue

            product = entry['product']
            if 'source' in product:
                product.pop('source')

            item = data.get(product['id'], None)
            if not item:
                result['failed_cnt'] += 1
                continue

            product.update(item)
            product_update_batch['entries'].append({
                'batchId': product_update_entry_idx,
                'merchantId': self.merchant_id,
                'method': 'insert',
                'product': product
            })
            product_update_entry_idx += 1

        if not product_update_batch['entries']:
            return result

        req = self.service.products().custombatch(body=product_update_batch)
        resp = req.execute(num_retries=self.max_retries)
        if resp['kind'] != 'content#productsCustomBatchResponse':
            result['failed_cnt'] += len(resp['entries'])
            logger.warning("Unknown kind - %s", resp['kind'])
            return

        for entry in resp['entries']:
            if 'errors' in entry and entry['errors']:
                result['failed_cnt'] += 1

                if 'product' in entry and entry['product']:
                    logger.error(
                        'Product update failed! OfferId: %s, errors: %s',
                        entry['product']['offerId'], repr(entry['errors']))
                else:
                    logger.warning(
                        'Product update failed! Entry: %s', repr(entry))
            else:
                result['succeed_cnt'] += 1

                if 'product' in entry and entry['product']:
                    logger.info('[ProductUpdated] OfferId %s!', entry['product']['offerId'])
                else:
                    logger.info('[ProductUpdated] Entry: %s,', repr(entry))

        return result

    def cleanup(self, issues_to_ignore=[], issues_to_record=['general_spam_policy', 'policy_violation']):
        sevirity_to_handle = ['critical', 'disapproved']
        issues_to_skip = ['price_mismatch', 'missing_required_gtin_error']
        issues_to_skip.extend(issues_to_ignore)

        if 'general_spam_policy' not in issues_to_record:
            issues_to_record.append('general_spam_policy')
        if 'policy_violation' not in issues_to_record:
            issues_to_record.append('policy_violation')

        page_size = 250
        request = self.service.productstatuses().list(
            merchantId=self.merchant_id, maxResults=page_size)

        i = 0
        result = {
            'succeed_cnt': 0,
            'failed_cnt': 0,
            'deleted': [],
            'spam': [],
        }
        while request is not None:
            resp = request.execute(num_retries=self.max_retries)
            request = self.service.productstatuses().list_next(request, resp)

            i += 1
            logger.debug("Downloaded merchant product status page %d" % i)

            prodstatus_buf = []
            if 'resources' in resp and resp['resources']:
                for prodstatus in resp['resources']:
                    if 'destinationStatuses' not in prodstatus:
                        logger.debug("No destinationStatuses found: %s", repr(prodstatus))
                        continue

                    shopping_stat = next(
                        (dest_stat for dest_stat in prodstatus['destinationStatuses'] \
                            if dest_stat['destination'] == 'Shopping'), None)
                    status = shopping_stat.get('approvalStatus', shopping_stat.get('status', None)) \
                        if shopping_stat else None
                    if status is None:
                        logger.warning('[MissingStatus] %s', prodstatus)
                        continue
                    if status != 'disapproved':
                        logger.debug("Ingore normal products: %s", repr(prodstatus))
                        continue

                    if 'dataQualityIssues' not in prodstatus and \
                        'itemLevelIssues' not in prodstatus:
                        logger.debug(
                            'Neither dataQualityIssues nor itemLevelIssues find in product,'
                            ' skip: %s',
                            repr(prodstatus))
                        continue

                    data_quality_issues = prodstatus.get('dataQualityIssues', [])
                    issue_ids = [data_quality_issue['id'] for data_quality_issue in \
                        data_quality_issues if data_quality_issue['severity'] in sevirity_to_handle]
                    item_level_issues = prodstatus.get('itemLevelIssues', [])
                    issue_ids.extend(
                        [item_level_issue['code'] for item_level_issue in item_level_issues \
                        if item_level_issue['servability'] in sevirity_to_handle])
                    if len(issue_ids) <= 0:
                        logger.debug(
                            "Ingore issues whose severity not in %s", ','.join(sevirity_to_handle))
                        continue

                    # Remove skippable issues
                    for issue_id in issues_to_skip:
                        if issue_id in issue_ids:
                            logger.debug("Ignore issue %s", issue_id)
                            issue_ids.remove(issue_id)

                    if len(issue_ids) > 0:
                        prodstatus_buf.append(prodstatus)

                        result['deleted'].append({
                            'productId': prodstatus['productId'],
                            'issue_ids': issue_ids
                        })

                        record_issue_ids = [issue_id for issue_id in issue_ids \
                            if issue_id in issues_to_record or issue_id.find('spam') != -1 or \
                                issue_id.find('policy') != -1]
                        if len(record_issue_ids) > 0:
                            result['spam'].append({
                                'productId': prodstatus['productId'],
                                'issue_ids': record_issue_ids
                            })

                if len(prodstatus_buf) <= 0:
                    continue

                prod_ids = [prod_stat['productId'] for prod_stat in prodstatus_buf]
                delete_result = self.delete_products(prod_ids)
                logger.info(
                    '[ProductDeletion] Succeed: %s, Failed: %s',
                    delete_result['succeed_cnt'], delete_result['failed_cnt'])

                result['succeed_cnt'] += delete_result['succeed_cnt']
                result['failed_cnt'] += delete_result['failed_cnt']
            else:
                break

        return result

    def update_inventory(self, inv_buf):
        result = {'succeed_cnt': 0, 'failed_cnt': 0}
        resp = None
        max_retry = 6
        batch = {'entries': []}
        for idx, item in enumerate(inv_buf):
            productId = item.pop('productId')
            entry = {
                'batchId': idx,
                'merchantId': self.merchant_id,
                'method': 'update',
                'productId': productId,
                'product': item
            }
            batch['entries'].append(entry)

        while max_retry > 0:
            try:
                request = self.service.products().custombatch(body=batch)
                resp = request.execute(num_retries=self.max_retries)
                break
            except SocketError as e:
                time.sleep(max_retry)
                if e.errno != errno.ECONNRESET:
                    max_retry -= 1
                    continue

                self.service = self.authenticate(self.service_account, self.service_version)
            except Exception as e:
                logger.exception(e)
                max_retry -= 1
                continue

        if resp is None:
            logger.error("[BatchProductUpdateFailed] All request failed!")
            result['failed_cnt'] += len(batch['entries'])
        else:
            if resp['kind'] == 'content#productsCustomBatchResponse' and 'entries' in resp:
                for entry in resp['entries']:
                    errors = entry.get('errors')
                    if errors:
                        result['failed_cnt'] += 1
                        logger.error(repr(errors))
                    else:
                        result['succeed_cnt'] += 1
            else:
                result['failed_cnt'] += len(batch)
                logger.error("Product update error - %s", resp)

        # for idx, item in enumerate(inv_buf):
        #     entry = item.copy()
        #     entry['batchId'] = idx
        #     batch['entries'].append(entry)

        # while max_retry > 0:
        #     try:
        #         request = self.service.inventory().custombatch(body=batch)
        #         resp = request.execute(num_retries=self.max_retries)
        #         break
        #     except SocketError as e:
        #         time.sleep(max_retry)
        #         if e.errno != errno.ECONNRESET:
        #             max_retry -= 1
        #             continue

        #         self.service = self.authenticate(self.service_account, self.service_version)
        #     except Exception as e:
        #         logger.exception(e)
        #         max_retry -= 1
        #         continue

        # if resp is None:
        #     logger.error("[BatchInventoryUpdateFailed] All request failed!")
        #     result['failed_cnt'] += len(batch['entries'])
        # else:
        #     if resp['kind'] == 'content#inventoryCustomBatchResponse' and 'entries' in resp:
        #         for entry in resp['entries']:
        #             errors = entry.get('errors')
        #             if errors:
        #                 result['failed_cnt'] += 1
        #                 logger.error(repr(errors))
        #             else:
        #                 result['succeed_cnt'] += 1

        #                 prod = batch['entries'][entry['batchId']]
        #                 logger.info(
        #                     "[MerchantId: %s][ProductId: %s] availability - %s, price - %s",
        #                     prod['merchantId'], prod['productId'],
        #                     prod['inventory']['availability'],
        #                     str(prod['inventory'].get('price', '-')))
        #     else:
        #         result['failed_cnt'] += len(batch)
        #         logger.error("Inventory update error - %s", resp)

        return result

    def update_products(self, prod_buf):
        result = {'succeed_cnt': 0, 'failed_cnt': 0}
        resp = None
        max_retry = 6
        batch = {'entries': []}
        for idx, item in enumerate(prod_buf):
            productId = item.pop('productId')
            item.pop('id', None)
            item.pop('offerId', None)
            item.pop('source', None)
            entry = {
                'batchId': idx,
                'merchantId': self.merchant_id,
                'method': 'update',
                'productId': productId,
                'product': item
            }
            batch['entries'].append(entry)

        while max_retry > 0:
            try:
                request = self.service.products().custombatch(body=batch)
                resp = request.execute(num_retries=self.max_retries)
                break
            except SocketError as e:
                time.sleep(max_retry)
                if e.errno != errno.ECONNRESET:
                    max_retry -= 1
                    continue

                self.service = self.authenticate(self.service_account, self.service_version)
            except Exception as e:
                logger.exception(e)
                max_retry -= 1
                continue

        if resp is None:
            logger.error("[BatchProductUpdateFailed] All request failed!")
            result['failed_cnt'] += len(batch['entries'])
        else:
            if resp['kind'] == 'content#productsCustomBatchResponse' and 'entries' in resp:
                for entry in resp['entries']:
                    errors = entry.get('errors')
                    if errors:
                        result['failed_cnt'] += 1
                        logger.error(repr(errors))
                    else:
                        result['succeed_cnt'] += 1
            else:
                result['failed_cnt'] += len(batch)
                logger.error("Product update error - %s", resp)

        return result

    def upload_products(self, prod_buf):
        result = {
            'succeed_cnt': 0,
            'failed_cnt': 0
        }

        if not prod_buf:
            return result

        batch = {'entries': []}
        for idx, item in enumerate(prod_buf):
            entry = {
                'batchId': idx,
                'merchantId': self.merchant_id,
                'method': 'insert',
                'product': item
            }
            batch['entries'].append(entry)

        req = self.service.products().custombatch(body=batch)
        try:
            resp = req.execute(num_retries=self.max_retries)
        except Exception as e:
            logger.exception(e)
            return result
        if resp['kind'] != 'content#productsCustomBatchResponse':
            logger.warning("Unknown operation - %s", resp['kind'])
            return result

        for entry in resp['entries']:
            if 'errors' in entry and entry['errors']:
                result['failed_cnt'] += 1
                if 'product' in entry and entry['product']:
                    logger.error(
                        '[CreateProductFailed] OfferId: %s, errors: %s',
                        entry['product']['offerId'], repr(entry['errors']))
                else:
                    logger.warning(
                        '[CreateProductFailed] Entry: %s', repr(entry))
            else:
                result['succeed_cnt'] += 1
                if 'product' in entry and entry['product']:
                    logger.info('[ProductCreated] offerId %s!', entry['product']['offerId'])
                else:
                    logger.info(
                        '[ProductCreated] %s, but no product information return!', repr(entry))

        return result

    def delete_products(self, productids):
        result = {
            'succeed_cnt': 0,
            'failed_cnt': 0
        }
        batch = {'entries': []}
        for idx, prod_id in enumerate(productids):
            batch['entries'].append({
                'batchId': idx,
                'merchantId': self.merchant_id,
                'method': 'delete',
                'productId': prod_id
            })
        req = self.service.products().custombatch(body=batch)
        resp = req.execute(num_retries=self.max_retries)
        for entry in resp['entries']:
            if 'errors' in entry and entry['errors']:
                result['failed_cnt'] += 1
                logger.warning(
                    '[ProductDeleteFailed] ProductID: %s, Errors: %s',
                    batch['entries'][entry['batchId']]['productId'], entry['errors'])
            else:
                result['succeed_cnt'] += 1
                logger.info(
                    '[ProductDeleted] ProductID: %s',
                    batch['entries'][entry['batchId']]['productId'])

        return result

    def download_reports(self, query):
        req = self.service.reports().search(merchantId=self.merchant_id, body={'query': query})
        while req is not None:
            resp = req.execute(num_retries=self.max_retries)
            req = self.service.reports().search_next(req, resp)

            if 'results' not in resp or not resp['results']:
                break

            yield resp['results']

    def authenticate(self, service_account_info, service_version=SERVICE_VERSION):
        credentials = service_account.Credentials.from_service_account_info(
            service_account_info,
            scopes=[CONTENT_API_SCOPE])

        auth_http = google_auth_httplib2.AuthorizedHttp(
            credentials, http=http.set_user_agent(http.build_http(), APPLICATION_NAME))

        return discovery.build(SERVICE_NAME, service_version, http=auth_http)
