# -*- coding: utf-8 -*-

import socket
import time

try:
    import ssl
except ImportError:
    _ssl_SSLError = object()
else:
    _ssl_SSLError = ssl.SSLError

from wrapt import patch_function_wrapper
import google.cloud
# import google.cloud._http
from google.cloud import _http
from google.cloud.exceptions import (
    GoogleCloudError,
    Redirection,
    ClientError,
    BadRequest,
    Unauthorized,
    Forbidden,
    NotFound,
    MethodNotAllowed,
    Conflict,
    LengthRequired,
    PreconditionFailed,
    RequestRangeNotSatisfiable,
    TooManyRequests,
    ServerError,
    InternalServerError,
    MethodNotImplemented,
    BadGateway,
    ServiceUnavailable,
    GatewayTimeout
)

exceptions_to_retry = (
    TooManyRequests, InternalServerError, BadGateway, ServiceUnavailable, GatewayTimeout
)
exceptions_not_retry = (
    Redirection, BadRequest, Unauthorized, Forbidden, NotFound, MethodNotAllowed, Conflict,
    LengthRequired, PreconditionFailed, RequestRangeNotSatisfiable, MethodNotImplemented
)

@patch_function_wrapper(_http.JSONConnection, 'api_request')
def api_request(wrapped, instance, args, kwargs):
    val = None
    num_retries = 12
    max_retry = num_retries
    exception = None

    while num_retries > 0:
        try:
            val = wrapped(*args, **kwargs)
            break
        except exceptions_to_retry as e:
            exception = e
            num_retries -= 1

            time.sleep(max_retry - num_retries)
            continue
        except exceptions_not_retry as e:
            raise e
        except _ssl_SSLError as ssl_error:
            exception = ssl_error
        except socket.timeout as socket_timeout:
            # It's important that this be before socket.error as it's a subclass
            # socket.timeout has no errorcode
            exception = socket_timeout

            num_retries -= 1

            time.sleep(3)
            continue
        except socket.error as socket_error:
            # errno's contents differ by platform, so we have to match by name.
            if socket.errno.errorcode.get(socket_error.errno) not in {
                "WSAETIMEDOUT",
                "ETIMEDOUT",
                "EPIPE",
                "ECONNABORTED",
            }:
                raise

            exception = socket_error
            num_retries -= 1

            time.sleep(3)
            continue
        except Exception as e:
            raise e
    else:
        if exception:
            raise exception

    return val
